﻿using EPiServer.Core;
using EPiServer.Data;
using EPiServer.Data.Dynamic;
using EPiServer.Data.Dynamic.Providers;
using log4net;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Web;

namespace VppMigrationTool.Util
{
    class FindDdsConverter
    {
        private static ILog _log = LogManager.GetLogger(typeof(FindDdsConverter));
        ConverterContext _converterContext;

        internal FindDdsConverter(ConverterContext converterContext)
        {
            _converterContext = converterContext;
        }

        internal void Convert()
        {
            var fromStore = DynamicDataStoreFactory.Instance.GetStore("EPiServer.Find.Cms.UnifiedFileBestBetSelector");
            var toStore = DynamicDataStoreFactory.Instance.GetStore("EPiServer.Find.Cms.PageBestBetSelector");
            if (fromStore == null || toStore == null)
            {
                Info("No EPiServer Find stores to convert found");
                return;
            }

            Info("Converting EPiServer Find best bets for virtual paths");

            var db = EPiServer.ServiceLocation.ServiceLocator.Current.GetInstance<IDatabaseHandler>();
            var count = 0;
            foreach (var item in fromStore.ItemsAsPropertyBag().ToList())
            {
                var path = item["Path"] as string;
                if (String.IsNullOrEmpty(path))
                {
                    fromStore.Delete(item.Id);
                    Warn(String.Format("EPiServer Find: Failed to convert best bets Id={0} since path is missing", item.Id));
                    continue;
                }

                var contentLink = FindContentLinkFromPath(db, path);
                if (String.IsNullOrEmpty(contentLink))
                {
                    contentLink = FindContentLinkFromPath(db, HttpUtility.UrlDecode(path));
                    if (String.IsNullOrEmpty(contentLink))
                    {
                        Warn(String.Format("EPiServer Find: Failed to convert best bets Id={0} for path={1} since no matching content was found", item.Id, path));
                        if (!_converterContext.ContinueOnError)
                        {
                            throw new Exception(String.Format("EPiServer Find: Failed to find a media item for virtual path {0} when converting best bets", path));
                        }
                        fromStore.Delete(item.Id);
                        continue;
                    }
                }

                try
                {
                    var x = new PropertyBag();
                    x["PageLink"] = new PageReference(contentLink);
                    var newItem = toStore.Save(x);

                    db.Execute(() =>
                        {
                            using(var cmd = db.CreateCommand("UPDATE [dbo].[tblBigTableReference] SET RefIdValue=@p0, ElementType='EPiServer.Find.Cms.PageBestBetSelector, EPiServer.Find.Cms, Version=7.0.586.0, Culture=neutral, PublicKeyToken=8fe83dea738b45b7', ElementStoreName='EPiServer.Find.Cms.PageBestBetSelector' WHERE ElementStoreName='EPiServer.Find.Cms.UnifiedFileBestBetSelector' AND RefIdValue=@p1", System.Data.CommandType.Text,  newItem.StoreId, item.Id.StoreId))
                            {
                                cmd.ExecuteNonQuery();
                            }
                        });


                    fromStore.Delete(item.Id);
                    count++;
                }
                catch (Exception e)
                {
                    Error(String.Format("EPiServer Find: Failed to convert best bets Id={0} for path={1} to content={2}", item.Id, path, contentLink), e);
                    if (!_converterContext.ContinueOnError)
                    {
                        throw;
                    }
                }
            }

            DataStoreProvider.CreateInstance().DeleteStoreDefinition(fromStore.Name, true);

            Info(String.Format("Successfully converted {0} best bets for virtual paths", count));
        }

        private void Info(string message)
        {
            _log.Info(message);
            if (_converterContext.StatusEventHandler != null)
            {
                _converterContext.StatusEventHandler(new StatusEventArgs(State.Info, message, _converterContext.ConvertedItemInfo) { });
            }
        }

        private void Warn(string message)
        {
            _log.Warn(message);
            if (_converterContext.StatusEventHandler != null)
            {
                _converterContext.StatusEventHandler(new StatusEventArgs(State.Warn, message, _converterContext.ConvertedItemInfo) { });
            }
        }

        private void Error(string message, Exception e)
        {
            _log.Error(message, e);
            if (_converterContext.StatusEventHandler != null)
            {
                _converterContext.StatusEventHandler(new StatusEventArgs(State.Error, message + " [" + e.Message + "]", _converterContext.ConvertedItemInfo) { StackTrace = e.StackTrace });
            }
        }

        private static string FindContentLinkFromPath(IDatabaseHandler db, string path)
        {
            return db.Execute(() =>
            {
                using (var cmd = db.CreateCommand("SELECT ContentLink FROM [dbo].[_MigratedVPPFiles] WHERE VirtualPath=@p0", System.Data.CommandType.Text, path))
                {
                    return cmd.ExecuteScalar() as string;
                }
            });
        }
    }
}
