﻿using EPiServer;
using EPiServer.Configuration;
using EPiServer.Core;
using EPiServer.DataAbstraction;
using EPiServer.DataAbstraction.RuntimeModel;
using EPiServer.Framework.Configuration;
using EPiServer.Framework.Initialization;
using EPiServer.ServiceLocation;
using EPiServer.Web.Hosting;
using System;
using System.Collections.Generic;
using System.Configuration;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading;
using System.Web.Hosting;

namespace VppMigrationTool.Util
{
    [Serializable]
    public delegate void StatusEventHandler(StatusEventArgs e);

    public class SiteProxy : MarshalByRefObject
    {
        public void InitAssemblyResolve()
        {
            EPiServer.Framework.Validator.ThrowIfNull("Check if Framework can be loaded", "Framework cannot be loaded");
            AppDomain.CurrentDomain.AssemblyResolve += CurrentDomain_AssemblyResolve;
        }

        private void InitalizeEPiServer(string webConfigPath)
        {
            if (!File.Exists(webConfigPath))
            {
                throw new ArgumentException("The configuration file '" + webConfigPath + "' does not exist", "webConfigPath");
            }

            var fileMap = new ExeConfigurationFileMap { ExeConfigFilename = webConfigPath };
            var config = ConfigurationManager.OpenMappedExeConfiguration(fileMap, ConfigurationUserLevel.None);
            EPiServerSection.ConfigurationInstance = config;

            EPiServerFrameworkSection frameworkSection = config.GetSection("episerver.framework") as EPiServerFrameworkSection;
            VirtualPathHandler.SetBypassAccessCheck(frameworkSection.VirtualPathProviders);
            EPiServer.Framework.Configuration.EPiServerFrameworkSection.ConfigurationInstance = config;

            EPiServer.Configuration.Settings.InitializeAllSettings(config);

            var es = (EPiServerSection)config.GetSection("episerver");

            EPiServer.Framework.Initialization.InitializationModule.FrameworkInitialization(HostType.Installer);
        }
        private string InitializeHostingEnvironment(string destinationPath, string virtualDirectory)
        {
            if (string.IsNullOrEmpty(destinationPath)) throw new NullReferenceException("destinationPath");
            if (!Directory.Exists(destinationPath)) throw new ArgumentException("The directory '" + destinationPath + "' does not exist.", "destinationPath");

            if (string.IsNullOrEmpty(virtualDirectory))
            {
                virtualDirectory = "/";
            }


            var noneWebContextHostingEnvironment = new NoneWebContextHostingEnvironment();
            noneWebContextHostingEnvironment.ApplicationVirtualPath = virtualDirectory;// : virtualDirectory.Insert(0, "/");
            noneWebContextHostingEnvironment.ApplicationPhysicalPath = destinationPath;
            GenericHostingEnvironment.Instance = noneWebContextHostingEnvironment;

            var fallbackMapPathVPPConfig = new System.Collections.Specialized.NameValueCollection();
            fallbackMapPathVPPConfig.Add("physicalPath", AppDomain.CurrentDomain.BaseDirectory);
            fallbackMapPathVPPConfig.Add("virtualPath", "~/");
            VirtualPathNonUnifiedProvider fallbackVPP = new VirtualPathNonUnifiedProvider("fallbackMapPathVPP", fallbackMapPathVPPConfig);
            GenericHostingEnvironment.Instance.RegisterVirtualPathProvider(fallbackVPP);

            Global.BaseDirectory = destinationPath;
            return virtualDirectory;
        }

        public SiteProxy()
        {
        }
        public void Convert(ConverterContext converterContext)
        {
            ThreadPool.QueueUserWorkItem(new WaitCallback((o) =>
            {
                var vppConverter = new VppConverter(converterContext);
                vppConverter.Convert();
            }));
        }

        public IEnumerable<String> GetVppNames()
        {
            List<String> res = new List<string>();
            foreach (VirtualPathProvider vpp in VirtualPathHandler.Instance.VirtualPathProviders.Keys)
            {
                if (vpp is VirtualPathUnifiedProvider)
                {
                    res.Add((vpp as VirtualPathUnifiedProvider).ProviderName);
                }
            }
            return res;
        }
        public IEnumerable<String> GetRegiteredFileData()
        {
            var filedatas = ServiceLocator.Current.GetInstance<IContentTypeRepository>().List().Where(ct => typeof(MediaData).IsAssignableFrom(ct.ModelType));
            List<String> res = new List<string>();
            foreach (var fileData in filedatas)
            {
                res.Add(fileData.ModelType.FullName);
            }
            return res;
        }
        public IEnumerable<String> GetLanguages()
        {
            return ServiceLocator.Current.GetInstance<ILanguageBranchRepository>().ListEnabled().Select(b => b.Culture.Name).ToList();
        }
        public override object InitializeLifetimeService()
        {
            return null;
        }

        public void Initialize(string destinationPath, string virtualDirectory)
        {
            virtualDirectory = InitializeHostingEnvironment(destinationPath, virtualDirectory);

            // Do not do this in the constructor as it can throw causing the dispose pattern to fail on instances of this object
            string webConfigPath = Path.Combine(destinationPath, "web.config");
            InitalizeEPiServer(webConfigPath);
        }

        Assembly CurrentDomain_AssemblyResolve(object sender, ResolveEventArgs args)
        {
            if (args.Name.Contains("VppMigrationTool"))
            {
                return typeof(StatusHandler).Assembly;
            }


            var assemblyName = new AssemblyName(args.Name);
            var assemblies = AppDomain.CurrentDomain.GetAssemblies();
            foreach (var a in assemblies)
            {
                var theAsemblyName = a.GetName();
                var b = theAsemblyName.IsSameAssemblyName(assemblyName);
            }
            Assembly assembly = assemblies.Where(a => a.GetName().IsSameAssemblyName(assemblyName)).FirstOrDefault();

            if (assembly == null && Global.BaseDirectory != null)
            {
                string assemblyPath = Path.Combine(Global.BaseDirectory, "bin", assemblyName.Name + ".dll");
                if (File.Exists(assemblyPath))
                {
                    assembly = Assembly.LoadFile(assemblyPath);
                }
            }

            return null;
        }

    }

    public enum State
    {
        Info,
        Warn,
        Error
    };

    [Serializable]
    public class StatusEventArgs : EventArgs
    {
        public String Message { get; set; }
        public State State { get; set; }
        public ConvertedItemInfo ConvertedItemInfo { get; set; }
        public bool IsDone { get; set; }
        public string StackTrace { get; set; }

        public StatusEventArgs(State state, String message, ConvertedItemInfo convertedItemInfo)
        {
            Message = message;
            State = state;
            ConvertedItemInfo = convertedItemInfo;
        }
    }
}
