﻿using System.Collections.Concurrent;
using System.Threading;
using EPiServer;
using EPiServer.BaseLibrary.IO;
using EPiServer.Core;
using EPiServer.Data.Entity;
using EPiServer.DataAbstraction;
using EPiServer.DataAbstraction.RuntimeModel;
using EPiServer.Framework.Blobs;
using EPiServer.Security;
using EPiServer.ServiceLocation;
using EPiServer.Web;
using EPiServer.Web.Hosting;
using EPiServer.Web.Routing;
using log4net;
using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using System.Web.Hosting;
using EPiServer.DataAccess;
using System.Data;
using System.Data.SqlClient;
using EPiServer.Data;

namespace VppMigrationTool.Util
{
    public class VppConverter
    {
        private IContentRepository _contentRepository = ServiceLocator.Current.GetInstance<IContentRepository>();
        private ContentMediaResolver _contentMediaResolver = ServiceLocator.Current.GetInstance<ContentMediaResolver>();
        private IContentTypeRepository _contenttypeRepository = ServiceLocator.Current.GetInstance<IContentTypeRepository>();
        private BlobFactory _blobFactory = ServiceLocator.Current.GetInstance<BlobFactory>();
        private static ILog _log = LogManager.GetLogger(typeof(VppConverter));
        private ConverterContext _converterContext;
        private bool _stopProcessing = false;
        public int MaxDegreeOfParallelism = 1;
        public VppConverter(ConverterContext converterContext)
        {
            _converterContext = converterContext;
        }

        public void Convert()
        {
            HandleMultiThreadRunning();
            EnsureLogTableIsCreated();
            try
            {
                foreach (var folderRoot in GetAssetFolders())
                {
                    RebuildUrlSegment(folderRoot);
                }

                var providers = _converterContext.ProviderNames.Select(name => FindVirtualPathUnifiedProvider(name));

                Parallel.ForEach(providers, new ParallelOptions() { MaxDegreeOfParallelism = MaxDegreeOfParallelism },
                    sourceProvider => ExecuteWithHandleException(sourceProvider, () =>
                    {
                        PermanentFileLinkMapper.Instance.ClearResolvers();
                        if (sourceProvider != null)
                        {
                            HandleInfo(String.Format("Converting VPP: {0}", sourceProvider.ProviderName));
                            if (sourceProvider != VirtualPathHandler.PageFolderProvider)
                            {
                                var destinationFolder = _contentRepository.Get<ContentFolder>(SiteDefinition.Current.GlobalAssetsRoot);
                                if (_converterContext.CreateRootFolderPerVpp)
                                {
                                    destinationFolder = CreateRootFolder(sourceProvider, destinationFolder);
                                }
                                DirectoryConverter(
                                    sourceProvider.GetDirectory(sourceProvider.VirtualPathRoot) as UnifiedDirectory,
                                    destinationFolder);
                            }
                            else
                            {
                                HandlePageFolderFiles(sourceProvider);
                            }

                            bool? status;
                            _converterContext.ConvertedItemInfo.ProviderStatus.TryGetValue(sourceProvider.ProviderName, out status);
                            if (!status.HasValue)
                            {
                                _converterContext.ConvertedItemInfo.ProviderStatus.AddOrUpdate(sourceProvider.ProviderName, true, (s, a) => true);
                            }
                        }
                    }));
                new SoftLinkUpdater().UpdateReferencedContent();
                new FindDdsConverter(_converterContext).Convert();
            }
            catch (Exception e)
            {
                HandleExcpetion(e, true);
            }
            finally
            {
                if (_converterContext.StatusEventHandler != null)
                {
                    _converterContext.StatusEventHandler(new StatusEventArgs(State.Info, "Done", _converterContext.ConvertedItemInfo.Copy()) { IsDone = true });
                }
            }
        }

        protected virtual void HandleMultiThreadRunning()
        {
            if (_converterContext.RunParallel && (_converterContext.CreateRootFolderPerVpp || _converterContext.ProviderNames.Count() == 1))
            {
                MaxDegreeOfParallelism = -1;
            }
        }

        protected virtual bool StopProcessing()
        {
            return _stopProcessing && !_converterContext.ContinueOnError;
        }

        private ContentFolder CreateRootFolder(VirtualPathUnifiedProvider sourceProvider, ContentFolder destinationFolder)
        {
            Interlocked.Increment(ref _converterContext.ConvertedItemInfo.NumberOfDirectory);

            UnifiedDirectory root = sourceProvider.GetDirectory(sourceProvider.VirtualPathRoot) as UnifiedDirectory;
            if (root == null)
            {
                throw new EPiServerException(String.Format("No Root folder for {0}",sourceProvider.VirtualPathRoot));
            }

            if (root.ACL != null && root.ACL.Entries.Count() == 1 && root.ACL.Entries.Single().Name == EveryoneRole.RoleName)
            {
                var inheritedAcl = root.ACL.CreateWritableClone();
                inheritedAcl.IsInherited = true;
                return CreateSubFolder<ContentFolder>(destinationFolder.ContentLink, root.Name, inheritedAcl, Guid.Empty);
            }
            return CreateSubFolder<ContentFolder>(destinationFolder.ContentLink, root.Name, root.ACL, Guid.Empty);
        }

   
        private HashSet<ContentReference> GetAssetFolders()
        {
            HashSet<ContentReference> assetFolders = new HashSet<ContentReference>();
            ServiceLocator.Current.GetInstance<SiteDefinitionRepository>().List().ToList()
                .ForEach(s =>
                    {
                        assetFolders.Add(s.GlobalAssetsRoot);
                        assetFolders.Add(s.SiteAssetsRoot);
                    });
            return assetFolders;
        }

        private void RebuildUrlSegment(ContentReference contentLink)
        {
            foreach (var folder in _contentRepository.GetChildren<ContentFolder>(contentLink))
            {
                ContentFolder routableContent = folder.CreateWritableClone() as ContentFolder;
                routableContent.RouteSegment = null;
                _contentRepository.Save(routableContent, EPiServer.DataAccess.SaveAction.ForceCurrentVersion | SaveAction.Publish | SaveAction.SkipValidation, AccessLevel.NoAccess);
                RebuildUrlSegment(folder.ContentLink);
            }
        }

        private void HandlePageFolderFiles(VirtualPathUnifiedProvider sourceProvider)
        {
            UnifiedDirectory rootDir = sourceProvider.GetDirectory(sourceProvider.VirtualPathRoot) as UnifiedDirectory;
            Parallel.ForEach(rootDir.GetDirectories().ToList<UnifiedDirectory>(), new ParallelOptions() { MaxDegreeOfParallelism = MaxDegreeOfParallelism }, unifiedDir =>
            {
                ExecuteWithHandleException(sourceProvider, () =>
                    {

                        Interlocked.Increment(ref _converterContext.ConvertedItemInfo.NumberOfDirectory);
                        IContent content = UnifiedDirectory.GetOwnerContentFromVirtualPath(unifiedDir.VirtualPath);
                        IResourceable contentResourceable = content as IResourceable;
                        if (contentResourceable != null)
                        {
                            var readOnly = contentResourceable as IReadOnly;
                            if (readOnly != null)
                            {
                                contentResourceable = readOnly.CreateWritableClone() as IResourceable;
                            }
                            var subFolder =
                                CreateSubFolder<ContentAssetFolder>(SiteDefinition.Current.ContentAssetsRoot,
                                                                    content.Name, null, content.ContentGuid);
                            contentResourceable.ContentAssetsID = subFolder.ContentGuid;
                            _contentRepository.Save(contentResourceable as IContent,
                                                    EPiServer.DataAccess.SaveAction.ForceCurrentVersion | EPiServer.DataAccess.SaveAction.SkipValidation |
                                                    EPiServer.DataAccess.SaveAction.Publish, AccessLevel.NoAccess);
                            DirectoryConverter(unifiedDir, subFolder);
                        }
                    });
            });
        }

        protected virtual void DirectoryConverter(UnifiedDirectory sourceDir, ContentFolder destinationFolder)
        {
            HandleInfo(String.Format("Converting files in {0} [{1}]", sourceDir.VirtualPath, destinationFolder.ContentLink));
            HandleUnifiedFiles(sourceDir.GetFiles(), destinationFolder);
            HandleInfo(String.Format("Converting folders in {0} [{1}]", sourceDir.VirtualPath, destinationFolder.ContentLink));
            HandleUnifiedDirectory(sourceDir.GetDirectories(), destinationFolder);
        }

        protected virtual void HandleUnifiedDirectory(UnifiedDirectory[] unifiedDirs, ContentFolder destinationContent)
        {
            Parallel.ForEach(unifiedDirs, new ParallelOptions() { MaxDegreeOfParallelism = MaxDegreeOfParallelism },
                unifiedDir => ExecuteWithHandleException(unifiedDir.Provider, () =>
                {
                    Interlocked.Increment(ref _converterContext.ConvertedItemInfo.NumberOfDirectory);

                    ContentFolder subfolder = CreateSubFolder<ContentFolder>(destinationContent.ContentLink,
                                                                             unifiedDir.Name, unifiedDir.ACL, Guid.Empty);
                    if (subfolder != null)
                    {
                        DirectoryConverter(unifiedDir, subfolder);
                    }
                }));
        }

        private void ExecuteWithHandleException(VirtualPathUnifiedProvider provider, Action a)
        {
            try
            {
                if (StopProcessing())
                {
                    return;
                }
                a();
            }
            catch (Exception e)
            {
                _converterContext.ConvertedItemInfo.ProviderStatus.AddOrUpdate(provider.ProviderName, false, (s, b) => false);
                HandleExcpetion(e);
            }
        }
        private void HandleACL(AccessControlList accessControlList, ContentFolder destinationContent)
        {
            if (accessControlList != null)
            {
                var descriptor = (destinationContent as IContentSecurable).GetContentSecurityDescriptor();
                if (accessControlList.IsInherited || destinationContent is IContentAsset)
                {
                    descriptor.IsInherited = true;
                }
                else
                {
                    descriptor.Clear();
                    foreach (KeyValuePair<string, AccessControlEntry> ace in accessControlList)
                    {
                        (destinationContent as IContentSecurable).GetContentSecurityDescriptor().AddEntry(ace.Value);
                    }
                }
                ServiceLocator.Current.GetInstance<IContentSecurityRepository>().Save(destinationContent.ContentLink, (destinationContent as IContentSecurable).GetContentSecurityDescriptor(), SecuritySaveType.Replace);
            }
        }

        protected virtual ContentFolder CreateSubFolder<T>(ContentReference parentReference, String folderName, AccessControlList accessControlList, Guid ownerGuid) where T : ContentFolder
        {
            try
            {
                UrlSegment.GetUrlFriendlySegment(folderName);
                var subDir = _contentRepository.GetBySegment(parentReference,
                                                             UrlSegment.GetUrlFriendlySegment(folderName),
                                                             NullLanguageSelector.Instance) as ContentFolder;
                if (subDir == null)
                {
                    HandleInfo(String.Format("Creating folder {0} in [{1}]", folderName, parentReference.ToString()));
                    subDir = _contentRepository.GetDefault<T>(parentReference, NullLanguageSelector.Instance);
                    subDir.Name = folderName;
                    if (ownerGuid != Guid.Empty)
                    {
                        ((IContentAsset)subDir).ContentOwnerID = ownerGuid;
                    }
                    var theFolderRef = _contentRepository.Save(subDir, EPiServer.DataAccess.SaveAction.Publish | EPiServer.DataAccess.SaveAction.SkipValidation, EPiServer.Security.AccessLevel.NoAccess);
                    Interlocked.Increment(ref _converterContext.ConvertedItemInfo.NumberOfContentFolder);
                    HandleACL(accessControlList, subDir);
                    subDir = _contentRepository.Get<ContentFolder>(theFolderRef);
                }
                return subDir;
            }
            catch (Exception e)
            {
                HandleExcpetion(e);
            }
            return null;
        }

        protected virtual void HandleUnifiedFiles(UnifiedFile[] files, ContentFolder parent)
        {
            Parallel.ForEach(files, new ParallelOptions() { MaxDegreeOfParallelism = MaxDegreeOfParallelism },
                unifiedFile => ExecuteWithHandleException(unifiedFile.Provider, () =>
                {
                    Interlocked.Increment(ref _converterContext.ConvertedItemInfo.NumberOfFile);
                    CreateFileContent(unifiedFile, parent, GetFirstMatching(unifiedFile));
                }));
        }

        protected virtual void CreateFileContent(UnifiedFile srcFile, ContentFolder targetDir, Type typeOfFileData)
        {
            try
            {
                Guid fileId = GetFileGuid(srcFile);
                MediaData fileData = null;
                if (fileId != Guid.Empty)
                {
                    try
                    {
                        //maybe we have already a contentfile with same id (We have done once migration tool)
                        var c = _contentRepository.Get<IContent>(fileId);
                        if (c != null)
                        {
                            HandleWarning(String.Format("{0} already exists [{1}]", srcFile.VirtualPath, c.ContentLink));
                            return;
                        }
                    }
                    catch (ContentNotFoundException)
                    {
                    }
                }

                HandleInfo(String.Format("Creating file {0} in [{1}]", srcFile.VirtualPath, targetDir.ContentLink));

                fileData = CreateContent(srcFile, targetDir, typeOfFileData, fileId, fileData);
                Interlocked.Increment(ref _converterContext.ConvertedItemInfo.NumberOfFileData);
                HandleBlob(srcFile, fileData);
                Interlocked.Increment(ref _converterContext.ConvertedItemInfo.NumberOfBlobContainer);
                HandleSummary(srcFile, fileData, typeOfFileData);
                _contentRepository.Save(fileData, EPiServer.DataAccess.SaveAction.Publish | EPiServer.DataAccess.SaveAction.SkipValidation, EPiServer.Security.AccessLevel.NoAccess);

                LogMigratedVppFileToDatabase(srcFile, fileData);
            }
            catch (Exception e)
            {
                HandleExcpetion(e);
            }

        }

        private static void LogMigratedVppFileToDatabase(UnifiedFile srcFile, MediaData fileData)
        {
            try
            {
                var db = EPiServer.ServiceLocation.ServiceLocator.Current.GetInstance<EPiServer.Data.IDatabaseHandler>();
                db.Execute(() =>
                {
                    using (var cmd = db.CreateCommand("INSERT INTO [dbo].[_MigratedVPPFiles](VirtualPath, ContentGuid, ContentLink) VALUES (@p0, @p1, @p2)", CommandType.Text, srcFile.VirtualPath, fileData.ContentGuid, fileData.ContentLink.ToReferenceWithoutVersion().ToString()))
                    {
                        cmd.ExecuteNonQuery();
                    }
                });
                return;
            }
            catch (SqlException)
            {
                //Should not exist a file with same virtual path but upon exception lets delete and retry
                var db = EPiServer.ServiceLocation.ServiceLocator.Current.GetInstance<EPiServer.Data.IDatabaseHandler>();
                db.Execute(() =>
                {
                    using (var cmd = db.CreateCommand("DELETE FROM [dbo].[_MigratedVPPFiles] WHERE VirtualPath=@p0", CommandType.Text, srcFile.VirtualPath))
                    {
                        cmd.ExecuteNonQuery();
                    }
                    using (var cmd = db.CreateCommand("INSERT INTO [dbo].[_MigratedVPPFiles](VirtualPath, ContentGuid, ContentLink) VALUES (@p0, @p1, @p2)", CommandType.Text, srcFile.VirtualPath, fileData.ContentGuid, fileData.ContentLink.ToReferenceWithoutVersion().ToString()))
                    {
                        cmd.ExecuteNonQuery();
                    }
                });
            }
        }

        #region CreateSqlLogTable
        private const string LogTableSql = @"IF NOT EXISTS (SELECT * FROM dbo.sysobjects WHERE id = OBJECT_ID(N'[dbo].[_MigratedVPPFiles]'))
BEGIN
	CREATE TABLE [dbo].[_MigratedVPPFiles] (
		[VirtualPath] [varchar] (450) NOT NULL,
		[ConvertedDate] [datetime] NOT NULL DEFAULT(getdate()),
		[ContentGuid] [uniqueidentifier] NOT NULL,
		[ContentLink] [varchar] (255) NOT NULL
	)
    CREATE INDEX [IX_VirtualPath] ON [dbo].[_MigratedVPPFiles]
    (
	    [VirtualPath] ASC
    )
END";
        private static void EnsureLogTableIsCreated()
        {
            var db = EPiServer.ServiceLocation.ServiceLocator.Current.GetInstance<EPiServer.Data.IDatabaseHandler>();
            db.Execute(() =>
            {
                using (var cmd = db.CreateCommand(LogTableSql, CommandType.Text))
                {
                    cmd.ExecuteNonQuery();
                }

            });
        }
        #endregion


        private MediaData CreateContent(UnifiedFile srcFile, ContentFolder targetDir, Type typeOfFileData, Guid fileId, MediaData fileData)
        {
            fileData = _contentRepository.GetDefault<IContent>(targetDir.ContentLink,
                _contenttypeRepository.Load(typeOfFileData).ID, NullLanguageSelector.Instance) as MediaData;
            fileData.Name = srcFile.Name;
            if (_converterContext.KeepIdentity && fileId != Guid.Empty)
            {
                fileData.ContentGuid = fileId;
            }
            return fileData;
        }

        private void HandleBlob(UnifiedFile srcFile, MediaData fileData)
        {
            var blob = _blobFactory.CreateBlob(fileData.BinaryDataContainer, srcFile.Extension);
            fileData.BinaryData = blob;
            using (Stream readStream = srcFile.Open(FileMode.Open, FileAccess.Read))
            {
                using (var writeStream = blob.OpenWrite())
                {
                    StreamConsumer.CopyToEnd(readStream, writeStream);
                }
            }
        }

        private Guid GetFileGuid(UnifiedFile srcFile)
        {
            IVersioningFile vf;
            if (srcFile.TryAsVersioningFile(out vf))
            {
                if (vf is VersioningFile)
                {
                    return ((VersioningFile)vf).Guid;
                }
            }

            IFileResolver fileResolver = srcFile.Provider as IFileResolver;
            if (fileResolver != null)
            {
                Guid fileGuid;
                fileResolver.ResolveFile(srcFile.VirtualPath, out fileGuid);
                return fileGuid;
            }

            return GetFileGuidByFilePath(srcFile);
        }

        private Guid GetFileGuidByFilePath(UnifiedFile srcFile)
        {
            IContent theContent = null;
            var segments = srcFile.VirtualPath.Substring(srcFile.Provider.VirtualPathRoot.Length).Split('/');
            ContentReference parentRef = SiteDefinition.Current.GlobalAssetsRoot;
            var startIndex = 0;
            if (srcFile.Provider == VirtualPathHandler.PageFolderProvider)
            {
                IContent content = UnifiedDirectory.GetOwnerContentFromVirtualPath(srcFile.VirtualPath);
                if (content != null)
                {
                    IResourceable contentResource = content as IResourceable;
                    if (contentResource != null)
                    {
                        var ownerContentFolder = _contentRepository.Get<ContentFolder>(contentResource.ContentAssetsID, NullLanguageSelector.Instance);
                        if (ownerContentFolder != null)
                        {
                            parentRef = ownerContentFolder.ContentLink;
                        }
                    }
                }
                startIndex = 1;
            }

            for (int i = startIndex; i < segments.Length; i++)
            {
                theContent = _contentRepository.GetBySegment(parentRef, UrlSegment.GetUrlFriendlySegment(segments[i]),
                                                             NullLanguageSelector.Instance);
                if (theContent == null)
                {
                    break;
                }
                parentRef = theContent.ContentLink;
            }

            if (theContent != null)
            {
                return theContent.ContentGuid;
            }

            return Guid.Empty;
        }

        protected virtual void HandleExcpetion(Exception e)
        {
            HandleExcpetion(e, _converterContext.ContinueOnError);
        }

        protected virtual void HandleExcpetion(Exception e, bool continueOnError)
        {
            _stopProcessing = true;
            if (_log.IsErrorEnabled)
            {
                _log.Error("Unhandled exception", e);
            }

            if (_converterContext.StatusEventHandler != null)
            {
                _converterContext.StatusEventHandler(new StatusEventArgs(State.Error, e.Message, _converterContext.ConvertedItemInfo) { StackTrace = e.StackTrace });
            }
            if (continueOnError)
            {
                return;
            }
            throw e;
        }

        protected virtual void HandleInfo(String message)
        {
            if (_log.IsInfoEnabled)
            {
                _log.Info(message);
            }

            if (_converterContext.StatusEventHandler != null)
            {
                _converterContext.StatusEventHandler(new StatusEventArgs(State.Info, message, _converterContext.ConvertedItemInfo.Copy()));
            }
        }

        protected virtual void HandleWarning(String message)
        {
            if (_log.IsWarnEnabled)
            {
                _log.Warn(message);
            }
            if (_converterContext.StatusEventHandler != null)
            {
                _converterContext.StatusEventHandler(new StatusEventArgs(State.Warn, message, _converterContext.ConvertedItemInfo.Copy()));
            }
        }

        protected virtual void HandleSummary(UnifiedFile srcFile, MediaData fileData, Type typeOfFileData)
        {
            if (!_converterContext.KeepSummary)
            {
                return;
            }
            Hashtable srcSummary = new Hashtable(srcFile.Summary.Dictionary);
            foreach (DictionaryEntry entry in srcSummary)
            {
                var prop = typeOfFileData.GetProperty(entry.Key.ToString(), BindingFlags.Public | BindingFlags.SetProperty | BindingFlags.Instance);
                if (prop != null)
                {
                    try
                    {
                        if (prop.Name == "Category")
                        {
                            continue;
                        }
                        // Should we handle Category prop?
                        prop.SetValue(fileData, entry.Value, null);
                    }
                    catch (Exception e)
                    {
                        HandleExcpetion(e);
                    }
                }
            }

        }

        protected virtual Type GetFirstMatching(UnifiedFile unifiedFile)
        {
            Type mediaType = null;
            if (!String.IsNullOrEmpty(unifiedFile.Extension))
            {
                mediaType = _contentMediaResolver.GetFirstMatching(unifiedFile.Extension);
            }
            return mediaType ?? _contentMediaResolver.GetFirstMatching("*");
        }


        protected virtual string GetRootName(VirtualPathUnifiedProvider sourceProvider)
        {
            List<String> directories = new List<string>();
            directories.AddRange(sourceProvider.VirtualPathRoot.Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries));
            return directories[0];
        }

        private VirtualPathUnifiedProvider FindVirtualPathUnifiedProvider(string providerName)
        {
            foreach (VirtualPathProvider vpp in VirtualPathHandler.Instance.VirtualPathProviders.Keys)
            {
                if (vpp is VirtualPathUnifiedProvider)
                {
                    if (String.Compare((vpp as VirtualPathUnifiedProvider).ProviderName, providerName, StringComparison.OrdinalIgnoreCase) == 0)
                    {
                        return vpp as VirtualPathUnifiedProvider;
                    }
                }
            }
            return null;
        }
    }

    [Serializable]
    public class ConvertedItemInfo
    {
        public ConvertedItemInfo()
            : this(Enumerable.Empty<string>())
        {
        }

        public ConvertedItemInfo(IEnumerable<string> providerNames)
        {
            foreach (var provider in providerNames)
            {
                ProviderStatus.AddOrUpdate(provider, (bool?)null, (s, b) => null);
            }
        }

        public int NumberOfFile;
        public int NumberOfDirectory;
        public int NumberOfContentFolder;
        public int NumberOfFileData;
        public int NumberOfBlobContainer;
        internal ConcurrentDictionary<string, bool?> ProviderStatus = new ConcurrentDictionary<string, bool?>();

        public ConvertedItemInfo Copy()
        {
            return new ConvertedItemInfo()
            {
                NumberOfBlobContainer = this.NumberOfBlobContainer,
                NumberOfContentFolder = this.NumberOfContentFolder,
                NumberOfDirectory = this.NumberOfDirectory,
                NumberOfFile = this.NumberOfFile,
                NumberOfFileData = this.NumberOfFileData,
                ProviderStatus = new ConcurrentDictionary<string, bool?>(this.ProviderStatus)
            };
        }
        public override string ToString()
        {
            return string.Format("Directories: {0}\r\nFiles: {1}\r\nContentFolders: {2}\r\nFileContent: {3}\r\nBlobs: {4}",
                     this.NumberOfDirectory, this.NumberOfFile, this.NumberOfContentFolder, this.NumberOfFileData, this.NumberOfBlobContainer);
        }

    }
}
